# -*- coding: utf-8 -*-
##############################################################################
#
# metacomp : a metacompiler for RLL(1) grammars
# Copyright (C)  Juan Miguel Vilar
#                    Universitat Jaume I, Castelló (Spain)
#
# This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#  Any questions regarding this software should be directed to:
#
#   Juan Miguel Vilar
#   Departament de Llenguatges i Sistemes Informàtics
#   Universitat Jaume I
#   E12071 Castellón (SPAIN)
#
#   email: jvilar@lsi.uji.es
#
###############################################################################
#
# File: minimization.py
#
#   Minimization of DFAs
#
# Implemented following:
#   "Re-describing an algorithm by Hopcroft"
#   Timo Knuutila
#   Theoretical Computer Science 250 (2001) 333-363
#

from automaton import Automaton

# Structures from the article
inv_head = None
inv_elts = None
cls_head = None
cls_elts = None
states = None

L = None
L_member = None

A = None
X = None

MaxClass = None


# Auxiliary structures

X2V = None # Mapping from X to the original vocabulary

# Classes

class InvHeader:
    def __init__(self, first = -1, size = 0):
        self.first = first
        self.size = size

    def __str__(self):
        return "(f: %s, s: %s)" % (self.first, self.size)

class ClassHeader:
    def __init__(self, first, size):
        self.first = first
        self.size = size
        self.counter = 0

    def __str__(self):
        return "(f: %s, s: %s, c: %s)" % (self.first, self.size, self.counter)

class StateInfo:
    def __init__(self, cls_of, idx_of):
        self.cls_of = cls_of
        self.idx_of = idx_of

    def __str__(self):
        return "(cl: %s, id: %s)" % (self.cls_of, self.idx_of)


def collect(x, C):
    suspects = []
    CH = cls_head[C]
    for i in range(CH.first, CH.first + CH.size):
        inv_x_a = inv_head[x][cls_elts[i]]
        for j in range(inv_x_a.first, inv_x_a.first + inv_x_a.size):
            b = inv_elts[x][j]
            B = states[b].cls_of
            if cls_head[B].counter == 0:
                suspects.append(B)
                cls_head[B].move = []
            cls_head[B].counter += 1
            cls_head[B].move.append(b)
    return suspects

# Functions

def refine(suspects):
    for B in suspects:
        if cls_head[B].counter < cls_head[B].size:
            Bp = split(B, cls_head[B].move)
            cls_head[Bp].counter = 0
            for x in X:
                if (B,x) in L:
                    L.append((Bp, x))
                else:
                    addBetter(B, Bp, x, L)
        cls_head[B].counter = 0

def addBetter(B1, B2, x, L):
    if cls_head[B1].size <= cls_head[B2].size:
        L.append((B1, x))
    else:
        L.append((B2, x))

def split(B, move):
    global MaxClass
    MaxClass += 1
    Bnew = MaxClass
    hole = cls_head[B].first
    for a in move:
        apos = states[a].idx_of
        b = cls_elts[hole]
        states[a].idx_of = hole
        states[b].idx_of = apos
        cls_elts[hole] = a
        cls_elts[apos] = b
        states[a].cls_of = Bnew
        hole += 1
    cls_head[Bnew].first = cls_head[B].first
    cls_head[B].first = hole
    cls_head[B].size -= len(move)
    cls_head[Bnew].size = len(move)
    return Bnew

def minimize(automaton):
    global MaxClass, cls_head, cls_elts, states, L
    automaton.nonOverlapping()
    initialize(automaton)
    MaxClass = 0
    cls_head = [ ClassHeader(0, len(A)) for a in A ]
    cls_elts = A[:]
    states = [ StateInfo(0, a) for a in A ]

#    ver("Comienzo:", [(MaxClass, "MaxClass"), (cls_head, "cls_head"), (cls_elts, "cls_elts"), (states, "states"), (L, "L")])

    split (0, final)
    L = []
    for x in X:
        addBetter(0, 1, x, L)

#    ver("Primer split:", [(MaxClass, "MaxClass"), (cls_head, "cls_head"), (cls_elts, "cls_elts"), (states, "states"), (L, "L")])
#    verExpandido()

    while len(L) != 0:
        C, x = L.pop()
        suspects = collect(x, C)
        refine(suspects)
#        ver("Nueva vuelta:", [(MaxClass, "MaxClass"), (cls_head, "cls_head"), (cls_elts, "cls_elts"), (states, "states"), (L, "L")])
#        verExpandido()
    result = recover(automaton)
    result.compactArcs()
    return result

def initialize(automaton):
    """Initialize the global variables with the representation
    of the automaton."""
    global X, X2V, A, A2S, S2A, final, inv_head, inv_elts, closing, trans
    V = list(automaton.vocabulary())
    X2V = V
    X = range(len(X2V))
    S = list(automaton.states)
    A2S = S
    A = range(len(A2S)+1) # Add a state to close the automaton
    S2A = dict (zip(S, A))
    closing = len(A2S)
    A2S.append(None)
    needsClose = False
    trans = [ [0] * len(A) for x in X ]
    for x,v in enumerate(V):
        for a,s in enumerate(S):
            e = automaton.getTransition(s, v)
            if e != None:
                trans[x][a] = S2A[e]
            else:
                trans[x][a] = closing
                needsClose = True
    if needsClose:
        for x in X:
            trans[x][closing] = closing
    else:
        A = A[:-1] # Eliminate the closing state
    inv_head = [ [0] * len(A) for _ in X ]
    inv_elts = [ [0] * len(A) for _ in X ]
    for x in X:
        t = [(trans[x][a], a) for a in A]
        t.sort()
        free = 0
        current = -1
        inv_head[x] = [ InvHeader() for _ in A ]
        for (e, a) in t:
            if e != current:
                inv_head[x][e].first = free
                current = e
            inv_elts[x][free] = a
            free += 1
            inv_head[x][e].size += 1
    final = [ S2A[s] for s in automaton.finals ]

#    ver("initialize", [(X, "X"), (X2V, "X2V"), (A, "A"), (A2S, "A2S"), (S2A, "S2A"), (final, "final"), (inv_head, "inv_head"), (inv_elts, "inv_elts"), (closing, "closing"), (trans, "trans")])


def verExpandido():
    visto = set()
    print "\n- Clases:"
    for cl, header in enumerate(cls_head[:MaxClass+1]):
        if header.size > 0:
            eltos = map(str, cls_elts[header.first: header.first+header.size])
            print "%s: %s" % (cl, ", ".join(eltos))
        else:
            print "%s: vacía" % cl
    print "- Pendientes: %s" % comoCadena([ "(%s, %s)" % (B, X2V[x]) for B,x in L ])


def ver(m, l):
    print "=============================="
    print m
    for (v,t) in l:
        print "%s: %s" % (t, comoCadena(v))

def comoCadena(v):
    if isinstance(v, list):
        return "[ %s ]" % ", ".join([ "%d: %s" % (i, comoCadena(c)) for i,c in enumerate(v)])
    else:
        return str(v)

def recover(original):
    """Construct an automaton from the data structure."""
    m = Automaton()
    for s, header in enumerate(cls_head):
        isClosing = False
        isInitial = False
        for i in range(header.first, header.first+header.size):
            f = A2S[cls_elts[i]]
            if f == original.initial:
                isInitial = True
            if f == None:
                isClosing = True
        if isClosing:
            continue
        item = None
        for i in range(header.first, header.first+header.size):
            fp = A2S[cls_elts[i]]
            ofp = A2S[fp]
            if item == None:
                item = original.items[ofp]
            else:
                item = item.joinMarks(original.items[ofp])
        f = cls_elts[header.first] # We use f as representant of the class
        of = A2S[f] # f in the original automaton
        m.addState(s, item)
        if original.isFinal(of):
            m.makeFinal(s)
        if isInitial:
            m.makeInitial(s)
        for x in X:
            d = trans[x][f]
            if d != closing:
                m.addTransition(s, X2V[x], states[d].cls_of)
    return m
